﻿/* OfficeTenant_Worker.cs
 * 
 * Copyright (C) 2017 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics;

namespace Popolo.BuildingOccupant
{

  public partial class OfficeTenant
  {

    /// <summary>事務所執務者</summary>
    [Serializable]
    public class Worker : ImmutableWorker
    {

      #region 列挙型定義

      /// <summary>職業種別</summary>
      public enum CategoryOfJob
      {
        /// <summary>一般平社員</summary>
        NoTitle,
        /// <summary>管理職</summary>
        Manager,
        /// <summary>役員</summary>
        Administrator
      }

      /// <summary>昼休み</summary>
      private enum LunchBreakTake
      {
        /// <summary>常に社外</summary>
        AlwaysOutGo = 0,
        /// <summary>常に社内</summary>
        NeverOutGo = 1,
        /// <summary>時に社外</summary>
        SomeTimeOutGo = 2
      }

      #endregion

      #region クラス変数

      /// <summary>出社モデルのパラメータ（平日）</summary>
      private static readonly double[] sttMdl = new double[] { -40.23, 32.61, 1.101, -0.452 };

      /// <summary>出社モデルのパラメータ（休日）</summary>
      private static readonly double[] sttMdl_Hol = new double[] { 24.45, 98.71, 1.273, -0.060 };

      /// <summary>退社モデルのパラメータ（平日:M20-M40）</summary>
      private static readonly double[] endMdl_M24 = new double[] { 82.3, 109.28, 1.529, 0.438 };

      /// <summary>退社モデルのパラメータ（平日:その他）</summary>
      private static readonly double[] endMdl = new double[] { 40.86, 72.69, 1.309, 0.286 };

      /// <summary>退社モデル（滞在時間）のパラメータ（休日）</summary>
      private static readonly double[] endMdl_Hol = new double[] { -73.75, 203.35, 1.465, 0.683 };

      /// <summary>昼休み開始モデルのパラメータ</summary>
      private static readonly double[] lncStMdl = new double[] { -0.337, 1.881, 0.502, 0.303 };

      /// <summary>昼休み終了モデルのパラメータ</summary>
      private static readonly double[] lncEdMdl = new double[] { -5.661, 74.93, 0.686, 0.905 };

      //// <summary>社内→社内の状態推移確率</summary>
      //private static readonly double[] tProbII = new double[] { 0.973, 0.979, 0.986, 0.952, 0.960, 0.991, 0.964, 0.973, 0.989, 0.976 };

      /// <summary>社外→社外の状態推移確率</summary>
      private static readonly double[] tProbOO = new double[] { 0.880, 0.773, 0.850, 0.765, 0.816, 0.905, 0.556, 0.417, 0.556, 0.710 };

      #endregion

      #region インスタンス変数

      /// <summary>一様乱数発生器</summary>
      private readonly MersenneTwister urGen;

      /// <summary>正規乱数発生器</summary>
      private readonly NormalRandom nrGen;

      /// <summary>執務者グループ:M20,M30,M40,M50,M60,F20,F30,F40,F50,F60</summary>
      private int group;

      /// <summary>徹夜作業を行う可能性のある執務者か否か</summary>
      private bool allNightWorkable;

      /// <summary>休日出勤を行う可能性のある執務者か否か</summary>
      private bool holidayWorkable;

      /// <summary>昼休みの取り方</summary>
      private LunchBreakTake lunchBreak;

      /// <summary>徹夜作業中か</summary>
      private bool allNightWorker = false;

      /// <summary>休日出勤中か</summary>
      private bool holidayWorker = false;

      /// <summary>昼休み外出時刻</summary>
      private DateTime lnchOutGoTime;

      /// <summary>昼休み帰社時刻</summary>
      private DateTime lnchComeBackTime;

      /// <summary>外出開始時刻リスト</summary>
      private List<DateTime> outGoTime = new List<DateTime>();

      /// <summary>帰社時刻リスト</summary>
      private List<DateTime> comeBackTime = new List<DateTime>();

      /// <summary>社内→社内の状態推移確率</summary>
      private double tProbII;

      #endregion

      #region プロパティ定義

      /// <summary>男性か否かを取得する</summary>
      public bool IsMale { get; private set; }

      /// <summary>職業種別を取得する</summary>
      public CategoryOfJob Job { get; private set; }

      /// <summary>非正規雇用か否か</summary>
      public bool IsPermanent { get; private set; }

      /// <summary>年齢を取得する</summary>
      public int Age { get; private set; }

      /// <summary>所属する事務所を取得する</summary>
      public ImmutableOfficeTenant Office { get; private set; }

      /// <summary>出社時刻を取得する</summary>
      public DateTime ArriveTime { get; private set; }

      /// <summary>退社時刻を取得する</summary>
      public DateTime LeaveTime { get; private set; }

      /// <summary>休暇取得中か否かを取得する</summary>
      public bool IsLOA { get; private set; }

      /// <summary>平均在室率[-]を取得する</summary>
      public double AverageIndoorRate { get; private set; }

      /// <summary>社内に滞在しているか否かを取得する</summary>
      public bool StayInOffice { get; private set; }

      #endregion

      #region コンストラクタ

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="office">所属するテナント</param>
      /// <param name="isMale">男性か否か</param>
      /// <param name="age">年齢</param>
      /// <param name="isPermanent">正社員か否か</param>
      /// <param name="indoorRate">平均在室率[-]</param>
      /// <param name="uRnd">一様乱数生成器</param>
      public Worker(OfficeTenant office, bool isMale, int age, bool isPermanent, double indoorRate, MersenneTwister uRnd)
      {
        urGen = uRnd;
        nrGen = new NormalRandom(uRnd);

        //職業種別を確率的に決定
        CategoryOfJob job;
        double[] mngRate = new double[] { 4 / 80d, 9 / 80d, 23 / 81d, 31 / 83d, 18 / 59d, 1 / 72d, 1 / 78d, 4 / 79d, 6 / 67d, 7 / 67d };  //管理職比率
        double[] admRate = new double[] { 1 / 80d, 0 / 80d, 4 / 81d, 11 / 83d, 22 / 59d, 1 / 72d, 0 / 78d, 0 / 79d, 4 / 67d, 13 / 67d };  //役員比率
        double rnd = nrGen.NextDouble();
        if (mngRate[group] < rnd) job = CategoryOfJob.Manager;
        else if (mngRate[group] + admRate[group] < rnd) job = CategoryOfJob.Administrator;
        else job = CategoryOfJob.NoTitle;

        initialize(office, isMale, age, isPermanent, job, indoorRate);
      }

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="office">所属するテナント</param>
      /// <param name="isMale">男性か否か</param>
      /// <param name="age">年齢</param>
      /// <param name="isPermanent">正社員か否か</param>
      /// <param name="job">職業種別</param>
      /// <param name="indoorRate">平均在室率[-]</param>
      /// <param name="nRnd">一様乱数生成器</param>
      public Worker(OfficeTenant office, bool isMale, int age, bool isPermanent, CategoryOfJob job, double indoorRate, MersenneTwister nRnd)
      {
        urGen = nRnd;
        nrGen = new NormalRandom(nRnd);
        initialize(office, isMale, age, isPermanent, job, indoorRate);
      }

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="office">所属するテナント</param>
      /// <param name="isMale">男性か否か</param>
      /// <param name="age">年齢</param>
      /// <param name="isPermanent">正社員か否か</param>
      /// <param name="job">職業種別</param>
      /// <param name="indoorRate">平均在室率[-]</param>
      private void initialize(OfficeTenant office, bool isMale, int age, bool isPermanent, CategoryOfJob job, double indoorRate)
      {
        Office = office;
        IsMale = isMale;
        Age = age;
        IsPermanent = isPermanent;
        Job = job;

        //世代と年齢によるグループ判定
        if (IsMale)
        {
          if (age < 30) group = 0;
          else if (30 <= age && age < 40) group = 1;
          else if (40 <= age && age < 50) group = 2;
          else if (50 <= age && age < 60) group = 3;
          else group = 4;
        }
        else
        {
          if (age < 30) group = 5;
          else if (30 <= age && age < 40) group = 6;
          else if (40 <= age && age < 50) group = 7;
          else if (50 <= age && age < 60) group = 8;
          else group = 9;
        }

        //平均在室率から状態推移確率を計算
        tProbII = tProbOO[group] * (1 / indoorRate - 1) + 2 - 1 / indoorRate;

        //徹夜作業と休日出勤可能性判定
        if (IsPermanent)
        {
          allNightWorkable = urGen.NextDouble() < 0.045;
          holidayWorkable = urGen.NextDouble() < 0.306;
        }
        else
        {
          switch (Job)
          {
            case CategoryOfJob.NoTitle:
              allNightWorkable = urGen.NextDouble() < 0.093;
              holidayWorkable = urGen.NextDouble() < 0.355;
              break;
            case CategoryOfJob.Manager:
              allNightWorkable = urGen.NextDouble() < 0.215;
              holidayWorkable = urGen.NextDouble() < 0.527;
              break;
            case CategoryOfJob.Administrator:
              allNightWorkable = urGen.NextDouble() < 0.088;
              holidayWorkable = urGen.NextDouble() < 0.471;
              break;
          }
        }

        //昼休みの取り方
        double[] aGoOut = new double[] { 0.176, 0.157, 0.179, 0.122, 0.221, 0.068, 0.070, 0.141, 0.179, 0.155 };  //常に外出
        double[] nGoOut = new double[] { 0.392, 0.446, 0.488, 0.451, 0.453, 0.534, 0.581, 0.500, 0.488, 0.452 };  //常に社内
        double bf = urGen.NextDouble();
        if (bf < nGoOut[group]) lunchBreak = LunchBreakTake.NeverOutGo;
        else if (bf < nGoOut[group] + aGoOut[group]) lunchBreak = LunchBreakTake.AlwaysOutGo;
        else lunchBreak = LunchBreakTake.SomeTimeOutGo;
      }

      #endregion

      #region インスタンスメソッド

      /// <summary>1日の執務スケジュールを更新する</summary>
      /// <param name="dTime">日にち</param>
      public void UpdateDailySchedule(DateTime dTime)
      {
        //休日出勤判定
        IsLOA = Office.IsHoliday(dTime);
        if (!IsLOA && urGen.NextDouble() < 0.035)
        {
          IsLOA = true;
          holidayWorker = false;
        }
        else holidayWorker = IsLOA && holidayWorkable && urGen.NextDouble() < 0.188;
        //休日出勤者の34.1%は平日相当の業務時間
        if (holidayWorker && 0.341 < urGen.NextDouble()) IsLOA = false;

        //出社時刻計算****************************************
        //前日徹夜作業の場合には出社時刻=0:00,翌日は非休日
        bool spendNight = allNightWorker; //一次保存
        if (allNightWorker)
        {
          ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0);
          IsLOA = false;
        }
        //休日の場合
        else if (IsLOA && !holidayWorker)
          ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0).AddDays(1);
        //平日出勤・休日出勤の場合
        else
        {
          double[] par;
          if (!IsLOA) par = sttMdl; //平日出勤
          else par = sttMdl_Hol;    //休日出勤
          double min = par[0] + par[1] * Math.Sinh((nrGen.NextDouble() - par[3]) / par[2]);
          ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, Office.StartHour, Office.StartMinute, 0).AddMinutes(min);
        }

        //退社時刻計算****************************************
        //当日徹夜作業の場合には退社時刻=0:00
        allNightWorker = (urGen.NextDouble() < 0.012 && allNightWorkable);
        if (allNightWorker) LeaveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0).AddDays(1);
        //休日の場合
        else if (IsLOA && !holidayWorker) LeaveTime = ArriveTime;
        else
        {
          //平日出勤
          if (!IsLOA)
          {
            double[] par;
            if (group == 0 || group == 1 || group == 2) par = endMdl_M24;
            else par = endMdl;
            double min = par[0] + par[1] * Math.Sinh((nrGen.NextDouble() - par[3]) / par[2]);
            LeaveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, Office.EndHour, Office.EndMinute, 0).AddMinutes(min);
          }
          //休日出勤
          else
          {
            double min = endMdl_Hol[0] + endMdl_Hol[1] * Math.Sinh((nrGen.NextDouble() - endMdl_Hol[3]) / endMdl_Hol[2]);
            min += 60 * (Office.EndHour - Office.StartHour) + (Office.EndMinute - Office.StartMinute);
            LeaveTime = ArriveTime.AddMinutes(min);
          }
          if (LeaveTime.CompareTo(ArriveTime) < 0) LeaveTime = ArriveTime;
        }

        //昼休み**********************************************
        //外出する場合
        if (lunchBreak == LunchBreakTake.AlwaysOutGo || urGen.NextDouble() < 0.488)
        {
          double min = lncStMdl[0] + lncStMdl[1] * Math.Sinh((nrGen.NextDouble() - lncStMdl[3]) / lncStMdl[2]);
          lnchOutGoTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, Office.LunchStartHour, Office.LunchStartMinute, 0).AddMinutes(min);

          double z = (nrGen.NextDouble() - lncEdMdl[3]) / lncEdMdl[2];
          min = (lncEdMdl[0] + lncEdMdl[1] * Math.Exp(z)) / (Math.Exp(z) + 1);
          min = Math.Max(0, 60 * (Office.LunchEndHour - Office.LunchStartHour) + (Office.LunchEndMinute - Office.LunchStartMinute) - min);
          lnchComeBackTime = lnchOutGoTime.AddMinutes(min);
        }
        //社内に留まる場合
        else lnchOutGoTime = lnchComeBackTime = ArriveTime;

        //その他一時外出**************************************
        outGoTime.Clear();
        comeBackTime.Clear();
        DateTime dtNow;
        if (spendNight) dtNow = new DateTime(dTime.Year, dTime.Month, dTime.Day, Office.StartHour, Office.StartMinute, 0); //前日泊まった執務者は定時より外出し得る
        else dtNow = ArriveTime;
        bool goOut = urGen.NextDouble() < ((1.0 - tProbII) / (2.0 - tProbOO[group] - tProbII));
        if (goOut) outGoTime.Add(dtNow);
        dtNow = dtNow.AddHours(1);
        while (dtNow.CompareTo(LeaveTime) < 0)
        {
          if (goOut && (urGen.NextDouble() < 1.0 - tProbOO[group]))
          {
            goOut = false;
            comeBackTime.Add(dtNow);
          }
          else if (!goOut && (urGen.NextDouble() < 1.0 - tProbII))
          {
            goOut = true;
            outGoTime.Add(dtNow);
          }
          dtNow = dtNow.AddHours(1);
        }
        if (goOut) comeBackTime.Add(LeaveTime);
      }

      /// <summary>社内に滞在しているか否かを更新する</summary>
      /// <param name="dTime">時刻</param>
      /// <returns>社内に滞在しているか否か</returns>
      public void UpdateStatus(DateTime dTime)
      {
        //出社退社時間外の場合は不在
        if (dTime.CompareTo(ArriveTime) < 0 || 0 < dTime.CompareTo(LeaveTime))
        {
          StayInOffice = false;
          return;
        }

        //昼休み滞在不在確認
        if (0 < dTime.CompareTo(lnchOutGoTime) && dTime.CompareTo(lnchComeBackTime) <= 0)
        {
          StayInOffice = false;
          return;
        }

        //その他一時外出確認
        for (int i = 0; i < outGoTime.Count; i++)
        {
          if (0 < dTime.CompareTo(outGoTime[i]) && dTime.CompareTo(comeBackTime[i]) <= 0)
          {
            StayInOffice = false;
            return;
          }
        }

        StayInOffice = true;
      }

      /// <summary>就業形式を設定する</summary>
      /// <param name="allNightWorkable">徹夜作業をし得るか否か</param>
      /// <param name="holidayWorkable">休日出勤をし得るか否か</param>
      public void SetWorkStyle(bool allNightWorkable, bool holidayWorkable)
      {
        this.allNightWorkable = allNightWorkable;
        this.holidayWorkable = holidayWorkable;
      }

      /// <summary>オフィスに滞在している時間数[min/day]を計算する</summary>
      /// <returns>オフィスに滞在している時間数[min/day]</returns>
      public double CalculateStayInOfficeMinutes()
      {
        double sum = (LeaveTime - ArriveTime).TotalMinutes;
        for (int i = 0; i < outGoTime.Count; i++)
          sum -= (comeBackTime[i] - outGoTime[i]).TotalMinutes;
        sum -= (lnchComeBackTime - lnchOutGoTime).TotalMinutes;  //昼食時間が重複する可能性はある
        return Math.Max(sum, 0);
      }

      #endregion

    }

    #region 読み取り専用interface

    /// <summary>読み取り専用の事務所執務者</summary>
    public interface ImmutableWorker
    {

      /// <summary>男性か否かを取得する</summary>
      bool IsMale { get; }

      /// <summary>雇用形態を取得する</summary>
      Worker.CategoryOfJob Job { get; }

      /// <summary>年齢を取得する</summary>
      int Age { get; }

      /// <summary>所属する事務所を取得する</summary>
      ImmutableOfficeTenant Office { get; }

      /// <summary>出社時刻を取得する</summary>
      DateTime ArriveTime { get; }

      /// <summary>退社時刻を取得する</summary>
      DateTime LeaveTime { get; }

      /// <summary>休暇取得中か否かを取得する</summary>
      bool IsLOA { get; }

      /// <summary>社内に滞在しているか否かを取得する</summary>
      bool StayInOffice { get; }

      /// <summary>社内に滞在しているか否かを更新する</summary>
      /// <param name="dTime">時刻</param>
      void UpdateStatus(DateTime dTime);

      /// <summary>オフィスに滞在している時間数[min/day]を計算する</summary>
      /// <returns>オフィスに滞在している時間数[min/day]</returns>
      double CalculateStayInOfficeMinutes();

    }

    #endregion

  }
}
